/*********************************************************************
 *   Copyright 2018, UCAR/Unidata
 *   See netcdf/COPYRIGHT file for copying and redistribution conditions.
 *********************************************************************/

/** @file d4chunk.h
 * @brief Public header for the DAP4 chunk-protocol decoder.
 *
 * The actual declarations are in ncd4.h (NCD4_dechunk, NCD4_infermode, etc.).
 * This header is a placeholder included by d4includes.h.
 * @author Dennis Heimbigner
 */

#ifndef D4CHUNK_H
#define D4CHUNK_H 1

#endif /*D4CHUNK_H*/
